; debug.asm - Debugging aids
;
; Copyright (C) 1999 Thomas Novelli
;	Retro is distributed under the terms of the GNU LGPL, version 2 or
;	greater. (For the text of the license, see the file COPYING.LIB)
;
; Note: The actual debugger is in ddr.asm. These are commands you can run
; from ddr.
;

%include "boot.mac"
%include "macros.mac"
%include "odb.mac"
%include "memory.inc"

extern	print, print_num_pad, base, print_hex, putc, getc


init:	ret


	global md
md:
;
; Memory dump
;
; Usage:
;	push	<start address>
;	push	<# of lines to dump>	(16 bytes per line)
;
	push	ebp
	mov	ebp, esp
	pusha
	mpush	[base]
	mov dword [base], 16

	mov	esi, [ebp+12]		;ESI = Starting address
	mov	ecx, [ebp+8]		;ECX = Lines to print
	xor	eax, eax
.loop
	callproc print_hex, esi		;Print address
	space
	push	esi
	call	.8
	callproc putc, '-'
	space
	call	.8
	pop	esi
	call	.ascii
	cr
	loop	.loop

	mpop	[base]
	popa
	leave
	ret

.8	push	ecx
	mov	ecx, 8
.l8	lodsb
	callproc print_num_pad, eax, 0x82
	space
	loop	.l8
	pop	ecx
	ret

.ascii	push	ecx
	space
	mov	ecx, 16
.10	lodsb
	cmp	al, 26
	jbe	.11
	cmp	al, 127
	jb	.12
.11	mov	al, '.'
.12	callproc putc, eax
	loop	.10
	pop	ecx
	ret



	global mdw
mdw:
;
; Memory dump word
;
; Usage:
;	push	<start address>
;	push	<# of lines to dump>	(16 bytes per line)
;
	push	ebp
	mov	ebp, esp
	pusha

	mov	ebx, [ebp+12]		;EBX = Starting address
	mov	ecx, [ebp+8]		;ECX = Lines to print
.loop
	callproc print_hex, ebx		;Print address
	space

	push	ecx
	mov	ecx, 4			;Number of dwords per line
.line
	mov	eax, [ebx]
	callproc print_hex, eax		;Print a dword
	space
	add	ebx, 4
	loop	.line

	pop	ecx
	cr
	loop	.loop

	popa
	leave
	ret



	global show
show:
;;
;; Input:
%define .top	ebp+2*4	; top-level object
;;
	callproc show_helper, [esp+1*4], 0
	ret



show_helper:
;;
;; Input:
%define .top	ebp+3*4	; top-level object
%define .lvl	ebp+2*4	; recursion level (for indentation)
;;
	push	ebp
	mov	ebp, esp
	pusha
	mov	esi, [.top]
	mov	ecx, [.lvl]

	or	ecx, ecx
	jnz	.1
	callproc print, .msg	; Only print header at top level
	mov byte [.lines], 2
	jmp short .2

.1	mov	ecx, [.lvl]
	or	ecx, ecx
	jz	.2
	shl	ecx, 1
.indent	callproc putc, 32
	loop	.indent

.2	callproc print_hex,	esi
	space
	callproc print_hex,	[esi+ODB_parent]
	callproc print_hex,	[esi+ODB_next]
	callproc print_hex,	[esi+ODB_child]
	callproc print_hex,	[esi+ODB_addr]
	callproc print_hex,	[esi+ODB_flags]
	callproc print,		[esi+ODB_name]
	cr

	inc byte [.lines]		;Pause every screenful
	cmp byte [.lines], 24
	jb	.3
	callproc print, .more
	mov	[.lines], byte 0
	call	getc			;wait for keypress
	cr
	cmp	al, 27
	jne	.3
	mov dword [.lvl], 0		;Quit
	jmp short .done

.3	mov	ebx, [esi+ODB_child]	; show children
	or	ebx, ebx
	jz	.no_children
	mov	ecx, [.lvl]
	inc	ecx
	mpush	ebx, ecx	; recurse
	call	show_helper
	add	esp, 2*4
.no_children
	mov	ecx, [.lvl]
	or	ecx, ecx	; Don't show next object when reclvl=0
	jz	.done
	mov	esi, [esi+ODB_next]	; Next?
	or	esi, esi
	jnz near .1
.done
	popa
	leave
	ret

.msg	db	"Header    Parent   Next     Child    Address  Flags    Name",10
	db	"--------  -------- -------- -------- -------- -------- ----------",10,0
.lines	db	0
.more	db	"-- More (ESC to quit) --",0

