; odb.inc - Object Database
;
; Copyright (C) 1999 Thomas Novelli
;	Retro is distributed under the terms of the GNU LGPL, version 2 or
;	greater. (For the text of the license, see the file COPYING.LIB)
;
; A step toward persistent storage.. still compatible with RDF :)
;
; To do:
;	Handle paths, like "/forth/print", "io/write", etc.
;	Use '/', ':', or what?
;	Well, NASM allows '.' in identifiers, so use that here.
;		-done (sort of)
;
;	Non-hierachical organization: Objects with parent=0 could be treated
;	this way.  Any object with a parent is part of a hierarchy.  Don't
;	need any special flags.
;

%include "odb.mac"

%define IFS	'.'	; path separator


init_odb:
	;; Build 'root' object
	mov	edi, K_DATA
	mov	[ODB], edi	; [ODB] = head
	xor	eax, eax
	stosd			; next=0
	stosd			; child=0
	stosd			; parent=0
	stosd			; address=0
	stosd			; length=0
	mov	eax, T_OBJ
	stosd			; flags
	mov	eax, Nroot
	stosd			; name (string ptr)
;	xor	eax, eax
;	stosd			; extaddr=0

	mov	[DATA_end], edi
	mov	[CODE_end],	dword K_CODE

	;; Add other core objects
	;;
	;;               Name   	Parent	Address		Length	Flags
	;;		 ----		------	-------		------	-----
	callproc insert, Nkern,		K_DATA,	0,		0,	T_OBJ
	mov	eax, [esp-5*4]
	mov	[kern_hdr], eax
	callproc insert, Ninsert,	eax,	insert,		0,	T_FN
	callproc insert, Nsearch,	eax,	search,		0,	T_FN
	callproc insert, Nisearch,	eax,	isearch,	0,	T_FN
	callproc insert, Nchflags,	eax,	chflags,	0,	T_FN
	callproc insert, Npatch_int,	eax,	patch_int,	0,	T_FN
	callproc insert, Nprint,	eax,	print,		0,	T_FN
	callproc insert, Ncprint,	eax,	cprint,		0,	T_FN
	callproc insert, Nprint_hex,	eax,	print_hex,	0,	T_FN
	callproc insert, Nputc,		eax,	putc,		0,	T_FN
	callproc insert, Nattr,		eax,	attr,		1,	T_INT
	ret



	Nroot		string "root"
	Nkern		string "kern"
	Ninsert		string "insert"
	Nsearch		string "search"
	Nisearch	string "isearch"
	Nchflags	string "chflags"
	Npatch_int	string "patch_int"
	Nprint		string "print"
	Ncprint		string "cprint"
	Nprint_hex	string "print_hex"
	Nputc		string "putc"
	Nattr		string "attr"



;; Address of "kern" object, for later reference
kern_hdr	dd	0



insert:
;;
;; Input:
%define .name	ebp+6*4	; asciiz string
%define .parent	ebp+5*4	; object to insert under
%define .addr	ebp+4*4	; address of object to insert
%define .size	ebp+3*4	; size of object (0 if unknown)
%define	.flags	ebp+2*4	; type & flags
;;
;; Output:
%define .ptr	ebp+2*4	; Pointer to new ODB entry
;;
	push	ebp
	mov	ebp, esp
	pusha
	mov	edi, [DATA_end]
	push	edi
	mov	ebx, [.parent]
	mov	eax, [ebx+ODB_child]
	mov	[ebx+ODB_child], edi
	stosd	;; next = parent's previous "first child"
	xor	eax, eax
	stosd	;; child = 0
	mov	eax, [.parent]
	stosd
	mov	eax, [.addr]
	stosd
	mov	eax, [.size]
	stosd
	mov	eax, [.flags]
	stosd
	mov	eax, [.name]
	stosd
;	xor	eax, eax	;extaddr=0
;	stosd
	mov	[DATA_end], edi
	pop	edi
	mov	[.ptr], edi
	popa
	leave
	ret


chflags:
;;
;; Change the flags of an object (OR-mask)
;;
;; Input:
%define	.ptr	ebp+3*4	; ODB entry to change
%define	.flags	ebp+2*4 ; New flags (OR-mask)
;;
	push	ebp
	mov	ebp, esp
	pusha

	mov	eax, [.flags]
	mov	ebx, [.ptr]
	or	[ebx+ODB_flags], eax

	popa
	leave
	ret


%if 0		;; Counted string version
search:
;;
;; Look up an object by name
;;
;; Todo:
;;	- Return address of ODB entry
;;	- isearch (case-insensitive)
;;
;; Input:
%define .name	ebp+4*4	; asciiz string
%define .len	ebp+3*4	; length of name
%define .top	ebp+2*4	; top level to search under
;;
;; Output:
%define .flags	ebp+3*4 ; Flags of target object (??? if not found)
%define	.addr	ebp+2*4	; Address of target object (0 if not found)
;;
	push	ebp
	mov	ebp, esp
	pusha
	mov	esi, [.name]
	mov	edx, [.len]
	mov	ebx, [.top]
	mov	ebx, [ebx+ODB_child]
	jmp short .start
.search	mov	ebx, [ebx+ODB_next]
.start
	or	ebx, ebx
	jz	.done
	mov	edi, [ebx+ODB_name]
	movzx	ecx, byte [edi]
	cmp	ecx, edx	;compare lengths
	jne	.search
	inc	edi		;compare strings
	mov	esi, [.name]
	cld
	repe	cmpsb
	jne	.search
	mov	esi, [ebx+ODB_flags]
	mov	ebx, [ebx+ODB_addr]
.done	mov	[.addr], ebx
	mov	[.flags], esi
	popa
	leave
	ret
%endif



search:		;; asciiz string version
;;
;; Search w/ paths (parse path separators in object name)
;;
;; Input:
%define .name	ebp+5*4	; asciiz string
%define .len	ebp+4*4	; length of name (-1 = only stop at null terminator)
%define .top	ebp+3*4	; node to search under
%define .ifs	ebp+2*4	; path separator (0 = no paths)
;;
;; Output:
%define .oid	ebp+4*4	; ODB entry of target object
%define .flags	ebp+3*4 ; Flags of target object (??? if not found)
%define	.addr	ebp+2*4	; Address of target object (0 if not found)
;;
	push	ebp
	mov	ebp, esp
	pusha
	mov	esi, [.name]
	mov	ecx, [.len]
	mov	ebx, [.top]
	xor	edx, edx	;index into [.name]
;pchr "'"
;pstr esi
;pchr "'"
;dspace
;phex ecx

	;; Check ifs and first char...
	mov	al, [.ifs]
	or	al, al			; skip this if ifs=0
	jz	.presearch
	cmp	[esi], al		; If first char is separator, start
	jne	.presearch		; at the root object
	mov	ebx, [ODB]
	inc	edx
.presearch
;pchr '>'
	;; ebx = node to search under
	add	ebx, ODB_child - ODB_next
	mov	[.pos], edx

.search	;; main loop
;pchr '|'
	mov	ebx, [ebx+ODB_next]	; next object
	or	ebx, ebx
	jz	.done			; not found, so return 0
	mov	edi, [ebx+ODB_name]
	mov	edx, [.pos]
	;; match all characters of object name, until 0 is reached
.cmp	mov	al, [edi]
;pchr eax
	inc	edi
	or	al, al		;at end of object name??
	jz	.match1		;possible match
	cmp	edx, ecx	;end of search string reached?
	je .search		;next...
	mov	ah, [esi+edx]
	inc	edx
	cmp	al, ah
	je	.cmp
	jmp short .search	;next...

	;; It's a match.. *if* [esi] is 0 or ifs, or edx=[.len]
.match1
;pchr '*'
	cmp	ecx, -1
	je	.1		;length not specified..
;;;	inc	edx	;might need it..
	cmp	ecx, edx		;at end, as specified in [.len] ?
	je	.found
.1	cmp byte [esi+edx], 0		;null terminator?
	jz	.found
	cmp	al, [.ifs]		;path separator?
	jne	.search			;Try next object...
	;; Matched an element of the path; now search under current object...
	inc	edx
	jmp short .presearch

.found
;pchr '$'
	mov	[.oid], ebx
	mov	esi, [ebx+ODB_flags]
	mov	ebx, [ebx+ODB_addr]
.done	mov	[.addr], ebx
	mov	[.flags], esi
;dcr
	popa
	leave
	ret
.pos	dd	0


isearch:	;; Case-insensitive search
	jmp	search
